@@ -4,7 +4,8 @@
 
 ;; Author: Milan Glacier <dev@milanglacier.com>
 ;; Maintainer: Milan Glacier <dev@milanglacier.com>
-;; Version: 0.1.1
+;; Package-Version: 0.2
+;; Package-Revision: v0.2-0-gb69db24ddbcd
 ;; URL: https://github.com/milanglacier/termint.el
 ;; Package-Requires: ((emacs "29.1"))
 
@@ -46,7 +47,7 @@
 
 ;; termint provides a flexible way to define and manage interactions
 ;; with REPLs and CLI apps running inside a terminal emulator backend
-;; within Emacs. It allows you to easily configure how Emacs
+;; within Emacs.  It allows you to easily configure how Emacs
 ;; communicates with different REPLs, leveraging the capabilities of
 ;; fully-featured terminal emulators like `term', `vterm', or `eat'.
 
@@ -69,7 +70,9 @@
   :group 'tools)
 
 (defcustom termint-backend 'term
-  "The backend to use for REPL sessions."
+  "The backend to use for REPL sessions.
+Supported backends include `eat', `vterm', and the built-in `term'.
+Note that `eat' and `vterm' must be installed separately."
   :type '(choice (const :tag "eat" eat)
                  (const :tag "vterm" vterm)
                  (const :tag "term" term)))
@@ -107,11 +110,11 @@
 REPL-NAME is used to determine the buffer name, REPL-CMD is used to
 determine the shell command.  SESSION is a numeric suffix for the
 buffer name."
-  (let* ((repl-buffer-name (format "*%s*" repl-name))
-         (repl-shell (if (functionp repl-cmd)
-                         (funcall repl-cmd)
-                       repl-cmd)))
-    (pcase termint-backend
+  (let ((repl-buffer-name (format "*%s*" repl-name))
+        (repl-shell (if (functionp repl-cmd)
+                        (funcall repl-cmd)
+                      repl-cmd)))
+    (pcase-exhaustive termint-backend
       ('eat (termint--start-eat-backend repl-buffer-name repl-shell session))
       ('vterm (termint--start-vterm-backend repl-buffer-name repl-shell session))
       ('term
@@ -126,14 +129,14 @@ without a number is considered as session 0."
   (interactive)
   (when-let* ((buffer-name (prog1 (buffer-name)
                              (rename-buffer (concat (buffer-name) "--tmp"))))
-              (repl-name (when (string-match "^\\*\\(.*\\)\\*" buffer-name)
-                           (match-string 1 buffer-name)))
+              (repl-name (and (string-match "^\\*\\(.*\\)\\*" buffer-name)
+                              (match-string 1 buffer-name)))
               (sessions (seq-filter
                          #'get-buffer
-                         `(,(format "*%s*" repl-name)
-                           ,@(mapcar
-                              (lambda (x) (format "*%s*<%d>" repl-name x))
-                              (number-sequence 1 9))))))
+                         (cons (format "*%s*" repl-name)
+                               (mapcar
+                                (lambda (x) (format "*%s*<%d>" repl-name x))
+                                (number-sequence 1 9))))))
     (cl-loop for session in sessions
              for idx from 0
              do (with-current-buffer session
@@ -228,14 +231,12 @@ with REPL-NAME, initialized during each `termint-define' call."
          (bracketed-paste-start "\e[200~")
          (bracketed-paste-end "\e[201~")
          (string (funcall str-process-func string))
-         (start-pattern (if (stringp start-pattern) start-pattern
-                          (if multi-lines-p
-                              (plist-get start-pattern :multi-lines)
-                            (plist-get start-pattern :single-line))))
-         (end-pattern (if (stringp end-pattern) end-pattern
-                        (if multi-lines-p
-                            (plist-get end-pattern :multi-lines)
-                          (plist-get end-pattern :single-line))))
+         (start-pattern (cond ((stringp start-pattern) start-pattern)
+                              (multi-lines-p (plist-get start-pattern :multi-lines))
+                              (t (plist-get start-pattern :single-line))))
+         (end-pattern (cond ((stringp end-pattern) end-pattern)
+                            (multi-lines-p (plist-get end-pattern :multi-lines))
+                            (t (plist-get end-pattern :single-line))))
          (final-string
           (if multi-lines-p
               (concat start-pattern
@@ -317,15 +318,56 @@ line of that matches SOURCE-COMMAND."
   (save-excursion
     (if-let*
         ((beg (and
-               (funcall #'beginning-of-defun nil)
+               (beginning-of-defun)
                (point)))
          (end (progn
-                (funcall #'end-of-defun nil)
+                (end-of-defun)
                 (point))))
         (cons beg end)
       (message "No defun found at point")
       nil)))
 
+(defcustom termint-region-dispatchers
+  '(("paragraph" . termint--dispatch-paragraph)
+    ("buffer" . termint--dispatch-buffer)
+    ("defun" . termint--dispatch-defun))
+  "Alist mapping dispatchable region types to dispatcher functions.
+Each element is of the form (NAME . DISPATCHER).  The NAME is appended
+to generated command names (e.g., `termint-ipython-send-NAME' and
+`termint-ipython-source-NAME' for an ipython schema), while DISPATCHER
+should be a function returning a cons cell (BEG . END) for the
+corresponding region.  Customize this before calling `termint-define'
+to generate additional region commands automatically for each REPL
+schema."
+  :type '(alist :key-type string :value-type function))
+
+(defcustom termint-schema-custom-commands nil
+  "Alist of extra commands generated for each REPL schema.
+Each element takes the form (SUFFIX . FN).  SUFFIX is a string
+appended to `termint-REPL-NAME-` to create the final command name for
+every schema defined via `termint-define`.  FN is the function invoked
+by the generated command.  It receives two arguments: REPL-NAME and
+SESSION.  See `termint--hide-window` for a reference implementation of
+the expected function signature.  Configure this variable before
+invoking `termint-define` to ensure the custom commands are
+generated."
+  :type '(alist :key-type string :value-type function))
+
+(defvar termint-mode-map-additional-keys
+  '(("f" . "send-defun")
+    ("F" . "source-defun")
+    ("b" . "send-buffer")
+    ("B" . "source-buffer")
+    ("p" . "send-paragraph")
+    ("P" . "source-paragraph"))
+  "Alist of keys and command suffixes to add to generated REPL keymaps.
+Each element is a cons cell of the form (KEY . SUFFIX).  KEY is the
+string sequence to bind.  SUFFIX is a string appended to
+`termint-REPL-NAME-' to resolve the actual command symbol (e.g.,
+\"send-defun\" becomes `termint-ipython-send-defun' when defining an
+ipython REPL).  This variable should be modified before calling
+`termint-define' to affect the generated keymaps.")
+
 (defun termint--dispatch-region-and-send
     (dispatcher repl-name session source-syntax)
   "Get region via DISPATCHER, optionally transform for sourcing, and send.
@@ -480,15 +522,24 @@ enabled.  The default value is nil."
         (start-pattern-name (intern (concat "termint-" repl-name "-start-pattern")))
         (end-pattern-name (intern (concat "termint-" repl-name "-end-pattern")))
         (send-delayed-final-ret-name (intern (concat "termint-" repl-name "-send-delayed-final-ret")))
-        ;; send paragraph and source paragraph
-        (send-paragraph-func-name (intern (concat "termint-" repl-name "-send-paragraph")))
-        (source-paragraph-func-name (intern (concat "termint-" repl-name "-source-paragraph")))
-        ;; send buffer and source buffer
-        (send-buffer-func-name (intern (concat "termint-" repl-name "-send-buffer")))
-        (source-buffer-func-name (intern (concat "termint-" repl-name "-source-buffer")))
-        ;; send defun and source defun
-        (send-defun-func-name (intern (concat "termint-" repl-name "-send-defun")))
-        (source-defun-func-name (intern (concat "termint-" repl-name "-source-defun"))))
+        (region-definitions
+         (cl-loop for entry in termint-region-dispatchers
+                  for region-name = (car entry)
+                  for dispatcher = (cdr entry)
+                  for send-func-name = (intern (format "termint-%s-send-%s" repl-name region-name))
+                  for source-func-name = (intern (format "termint-%s-source-%s" repl-name region-name))
+                  collect (list :name region-name
+                                :dispatcher dispatcher
+                                :send send-func-name
+                                :source source-func-name)))
+        (custom-command-definitions
+         (cl-loop for entry in termint-schema-custom-commands
+                  for name = (car entry)
+                  for func = (cdr entry)
+                  for command-name = (intern (format "termint-%s-%s" repl-name name))
+                  collect (list :name name
+                                :func func
+                                :command command-name))))
 
     `(progn
 
@@ -565,62 +616,27 @@ process with that number."
                                ,str-process-func-name
                                ,send-delayed-final-ret-name))
 
-       (defun ,send-paragraph-func-name (&optional session)
-         ,(format
-           "Send the current paragraph to %s.
-With numeric prefix SESSION, send paragraph to the process associated
-with that number." repl-name)
-         (interactive "P")
-         (termint--dispatch-region-and-send
-          #'termint--dispatch-paragraph ,repl-name session nil))
-
-       (defun ,source-paragraph-func-name (&optional session)
-         ,(format
-           "Source the current paragraph to %s.
-With numeric prefix SESSION, send paragraph to the process associated
-with that number." repl-name)
-         (interactive "P")
-         (termint--dispatch-region-and-send
-          #'termint--dispatch-paragraph ,repl-name
-          session ,source-syntax-name))
-
-       (defun ,send-buffer-func-name (&optional session)
-         ,(format
-           "Send the current buffer to %s.
-With numeric prefix SESSION, send buffer to the process associated
-with that number." repl-name)
-         (interactive "P")
-         (termint--dispatch-region-and-send
-          #'termint--dispatch-buffer ,repl-name session nil))
-
-       (defun ,source-buffer-func-name (&optional session)
-         ,(format
-           "Source the current buffer to %s.
-With numeric prefix SESSION, send buffer to the process associated
-with that number." repl-name)
-         (interactive "P")
-         (termint--dispatch-region-and-send
-          #'termint--dispatch-buffer ,repl-name
-          session ,source-syntax-name))
-
-       (defun ,send-defun-func-name (&optional session)
-         ,(format
-           "Send the current defun to %s.
-With numeric prefix SESSION, send defun to the process associated
-with that number." repl-name)
-         (interactive "P")
-         (termint--dispatch-region-and-send
-          #'termint--dispatch-defun ,repl-name session nil))
-
-       (defun ,source-defun-func-name (&optional session)
-         ,(format
-           "Source the current defun to %s.
-With numeric prefix SESSION, send defun to the process associated
-with that number." repl-name)
-         (interactive "P")
-         (termint--dispatch-region-and-send
-          #'termint--dispatch-defun ,repl-name
-          session ,source-syntax-name))
+       ,@(cl-loop for entry in region-definitions append
+                  (list
+                   `(defun ,(plist-get entry :send) (&optional session)
+                      ,(format
+                        "Send the current %s to %s.
+With numeric prefix SESSION, send %s to the process associated
+with that number."
+                        (plist-get entry :name) repl-name (plist-get entry :name))
+                      (interactive "P")
+                      (termint--dispatch-region-and-send
+                       #',(plist-get entry :dispatcher) ,repl-name session nil))
+                   `(defun ,(plist-get entry :source) (&optional session)
+                      ,(format
+                        "Source the current %s to %s.
+With numeric prefix SESSION, send %s to the process associated
+with that number."
+                        (plist-get entry :name) repl-name (plist-get entry :name))
+                      (interactive "P")
+                      (termint--dispatch-region-and-send
+                       #',(plist-get entry :dispatcher) ,repl-name
+                       session ,source-syntax-name))))
 
        (when (require 'evil nil t)
          (evil-define-operator ,send-region-operator-name (beg end session)
@@ -649,22 +665,37 @@ suffix." repl-name)
          (interactive "P")
          (termint--hide-window ,repl-name session))
 
+       ,@(cl-loop for entry in custom-command-definitions append
+                  (let ((suffix (plist-get entry :name))
+                        (fn (plist-get entry :func))
+                        (command (plist-get entry :command)))
+                    (list
+                     `(defun ,command (&optional session)
+                        ,(format
+                          "Run custom command `%s' for %s.
+Generated from `termint-schema-custom-commands'.
+Calls `%S' with REPL-NAME and optional SESSION."
+                          suffix repl-name fn)
+                        (interactive "P")
+                        (funcall #',fn ,repl-name session)))))
+
        (defvar ,keymap-name
          (let ((map (make-sparse-keymap)))
            (define-key map "s" #',start-func-name)
            (define-key map "r" #',send-region-func-name)
            (define-key map "R" #',source-region-func-name)
            (define-key map "e" #',send-string-func-name)
-           (define-key map "p" #',send-paragraph-func-name)
-           (define-key map "f" #',send-defun-func-name)
-           (define-key map "P" #',source-paragraph-func-name)
-           (define-key map "b" #',send-buffer-func-name)
-           (define-key map "B" #',source-buffer-func-name)
-           (define-key map "F" #',source-defun-func-name)
            (define-key map "h" #',hide-window-func-name)
            (when (require 'evil nil t)
              (define-key map "r" #',send-region-operator-name)
              (define-key map "R" #',source-region-operator-name))
+           (dolist (binding termint-mode-map-additional-keys)
+             (let ((key (car binding))
+                   (suffix (cdr binding)))
+               (when-let* ((command (intern-soft (format "termint-%s-%s"
+                                                         ,repl-name suffix)))
+                           ((fboundp command)))
+                 (define-key map key command))))
            map)
          ,(format "Keymap for %s REPL commands." repl-name)))))
 
